@@ -157,27 +157,65 @@ class _GraphManager(object):
   
   def _get_annotation_axioms(self, source, property, target, target_d):
     if target_d is None:
-      for bnode in self._get_obj_triples_po_s(rdf_type, owl_axiom):
-        for p, o in self._get_obj_triples_s_po(bnode):
-          if   p == owl_annotatedsource: # SIC! If on a single if, elif are not appropriate.
-            if o != source: break
-          elif p == owl_annotatedproperty:
-            if o != property: break
-          elif p == owl_annotatedtarget:
-            if o != target: break
-        else:
-          yield bnode
+#       r = self.graph.execute("""
+# SELECT q1.s
+# FROM objs q1, objs q2 INDEXED BY index_objs_sp, objs q3 INDEXED BY index_objs_sp, objs q4 INDEXED BY index_objs_sp
+# WHERE q1.p=6 AND q1.o=?
+#   AND q2.s=q1.s AND q2.p=? AND q2.o=?
+#   AND q3.s=q1.s AND q3.p=? AND q3.o=?
+#   AND q4.s=q1.s AND q4.p=? AND q4.o=?""",
+#                          (owl_axiom,
+#                           owl_annotatedsource, source,
+#                           owl_annotatedproperty, property,
+#                           owl_annotatedtarget, target))
+#       for l in r.fetchall(): yield l[0]
+      r = self.graph.execute("""
+SELECT q1.s
+FROM objs q1, objs q2 INDEXED BY index_objs_sp, objs q3 INDEXED BY index_objs_sp, objs q4 INDEXED BY index_objs_sp
+WHERE q1.p=? AND q1.o=?
+  AND q2.s=q1.s AND q2.p=6 AND q2.o=?
+  AND q3.s=q1.s AND q3.p=? AND q3.o=?
+  AND q4.s=q1.s AND q4.p=? AND q4.o=?""",
+                         (owl_annotatedsource, source,
+                          owl_axiom,
+                          owl_annotatedproperty, property,
+                          owl_annotatedtarget, target))
+      for l in r.fetchall(): yield l[0]
+    
+      # for bnode in self._get_obj_triples_po_s(rdf_type, owl_axiom):
+      #   for p, o in self._get_obj_triples_s_po(bnode):
+      #     if   p == owl_annotatedsource: # SIC! If on a single if, elif are not appropriate.
+      #       if o != source: break
+      #     elif p == owl_annotatedproperty:
+      #       if o != property: break
+      #     elif p == owl_annotatedtarget:
+      #       if o != target: break
+      #   else:
+      #     yield bnode
     else:
-      for bnode in self._get_obj_triples_po_s(rdf_type, owl_axiom):
-        for p, o, d in self._get_triples_s_pod(bnode):
-          if   p == owl_annotatedsource: # SIC! If on a single if, elif are not appropriate.
-            if o != source: break
-          elif p == owl_annotatedproperty:
-            if o != property: break
-          elif p == owl_annotatedtarget:
-            if o != target: break
-        else:
-          yield bnode
+      r = self.graph.execute("""
+SELECT q1.s
+FROM objs q1, objs q2 INDEXED BY index_objs_sp, objs q3 INDEXED BY index_objs_sp, datas q4 INDEXED BY index_datas_sp
+WHERE q1.p=? AND q1.o=?
+  AND q2.s=q1.s AND q2.p=6 AND q2.o=?
+  AND q3.s=q1.s AND q3.p=? AND q3.o=?
+  AND q4.s=q1.s AND q4.p=? AND q4.o=?""",
+                         (owl_annotatedsource, source,
+                          owl_axiom,
+                          owl_annotatedproperty, property,
+                          owl_annotatedtarget, target))
+      for l in r.fetchall(): yield l[0]
+      
+      # for bnode in self._get_obj_triples_po_s(rdf_type, owl_axiom):
+      #   for p, o, d in self._get_triples_s_pod(bnode):
+      #     if   p == owl_annotatedsource: # SIC! If on a single if, elif are not appropriate.
+      #       if o != source: break
+      #     elif p == owl_annotatedproperty:
+      #       if o != property: break
+      #     elif p == owl_annotatedtarget:
+      #       if o != target: break
+      #   else:
+      #     yield bnode
           
   def _del_obj_triple_spo(self, s = None, p = None, o = None):
     #onto = CURRENT_NAMESPACES.get() or self
@@ -306,8 +344,9 @@ class _GraphManager(object):
         if   k == "iri":
           prop_vals.append((" iri", v, None))
         elif (k == "is_a") or (k == "subclass_of") or (k == "type"):
-          if isinstance(v, (_SearchMixin, Or)): v2 = v
-          else:                                 v2 = v.storid
+          if   isinstance(v, (_SearchMixin, Or)): v2 = v
+          elif isinstance(v, int):                v2 = v
+          else:                                   v2 = v.storid
           prop_vals.append((" %s" % k, v2, None))
         else:
           d = None
@@ -622,9 +661,12 @@ class World(_GraphManager):
     return entity
   
   def _parse_bnode(self, bnode):
-    for ontology in self.ontologies.values():
-      if ontology._has_obj_triple_spo(bnode, None, None):
-        return ontology._parse_bnode(bnode)
+    c = self.graph.db.execute("""SELECT c FROM objs WHERE s=? LIMIT 1""", (bnode,)).fetchone()
+    if c:
+      c = c[0]
+      for onto in self.ontologies.values():
+        if onto.graph.c == c: return onto._parse_bnode(bnode)
+      
       
      
 class Ontology(Namespace, _GraphManager):
@@ -935,10 +977,14 @@ class Ontology(Namespace, _GraphManager):
     
     with LOADING:
       restriction_property = restriction_type = restriction_cardinality = Disjoint = members = on_datatype = with_restriction = None
-      for pred, obj in self._get_obj_triples_s_po(bnode):
-        if   pred == owl_complementof:   r = Not(None, self, bnode); break # will parse the rest on demand
-        elif pred == owl_unionof:        r = Or (obj , self, bnode); break
-        elif pred == owl_intersectionof: r = And(obj , self, bnode); break
+      preds_objs = self._get_obj_triples_s_po(bnode)
+      if not preds_objs: # Probably a blank node from another ontology
+        return self.world._parse_bnode(bnode)
+      for pred, obj in preds_objs:
+        if   pred == owl_complementof:   r = Not          (None, self, bnode); break # will parse the rest on demand
+        elif pred == owl_unionof:        r = Or           (obj , self, bnode); break
+        elif pred == owl_intersectionof: r = And          (obj , self, bnode); break
+        #elif pred == owl_disjointunion:  r = DisjointUnion(obj , self, bnode); break
         
         elif pred == owl_onproperty: restriction_property = self._to_python(obj, None)
         
@@ -946,12 +992,6 @@ class Ontology(Namespace, _GraphManager):
         elif pred == ONLY:      restriction_type = ONLY;
         elif pred == VALUE:     restriction_type = VALUE;
         elif pred == HAS_SELF:  restriction_type = HAS_SELF;
-#        elif pred == EXACTLY:   restriction_type = EXACTLY; restriction_cardinality = self._to_python(obj, XXX)
-#        elif pred == MIN:       restriction_type = MIN;     restriction_cardinality = self._to_python(obj)
-#        elif pred == MAX:       restriction_type = MAX;     restriction_cardinality = self._to_python(obj)
-#        elif pred == owl_cardinality:     restriction_type = EXACTLY; restriction_cardinality = self._to_python(obj)
-#        elif pred == owl_min_cardinality: restriction_type = MIN;     restriction_cardinality = self._to_python(obj)
-#        elif pred == owl_max_cardinality: restriction_type = MAX;     restriction_cardinality = self._to_python(obj)
         
         elif pred == owl_oneof: r = OneOf(self._parse_list(obj), self, bnode); break
         
@@ -994,9 +1034,10 @@ class Ontology(Namespace, _GraphManager):
             #else:
             #  s = ""
             #  raise ValueError("Cannot parse blank node %s: unknown node type!")
+            
           else: # Not a blank
             r = self.world._get_by_storid(bnode, main_onto = self)
-            
+
     self._bnodes[bnode] = r
     return r
 
@@ -1040,6 +1081,24 @@ class Ontology(Namespace, _GraphManager):
         
   def __repr__(self): return """get_ontology("%s")""" % (self.base_iri)
   
+  def get_parents_of(self, entity):
+    if isinstance(entity, Thing):
+      t = self._get_obj_triples_sp_o(entity.storid, rdf_type)
+    else:
+      t = self._get_obj_triples_sp_o(entity.storid, entity._rdfs_is_a)
+    r = []
+    for o in t:
+      if o == owl_named_individual: continue
+      if o < 0: r.append(self._parse_bnode(o))
+      else:     r.append(self.world._get_by_storid(o))
+    return r
+  
+  def get_instances_of(self, Class):
+    return [self.world._get_by_storid(o) for o in self._get_obj_triples_po_s(rdf_type, Class.storid)]
+  
+  def get_children_of(self, Class):
+    return [self.world._get_by_storid(o) for o in self._get_obj_triples_po_s(Class._rdfs_is_a, Class.storid)]
+  
   
 class Metadata(object):
   def __init__(self, namespace, storid):
